/*
    lkpatch - a Linux Kernel Patcher

    Copyright 1999, 2000 Theo Nolte <nolte@post.rwth-aachen.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include <netinet/in.h>

#include "lkpatch.h"

#define BUFMAG 0x12435688

struct buf_t *newbuf(void) { /* return a new buffer */
 struct buf_t *bp;
 int sz=512;
  bp=emalloc(sizeof(struct buf_t));
  memset(bp,0,sizeof(bp));
  bp->magic=BUFMAG;
  bp->p=emalloc(sz);
  bp->sz=sz;
  bp->pos=bp->frozen=0;
  bp->file=NULL;
  bp->fd=-1;
  return(bp);
}

static void buffreeze(struct buf_t *bp) {
  if(bp->magic!=BUFMAG) error(1,0,"buffreeze: arg passed is not a buffer");
  if(!bp->frozen) {
    bp->frozen=1;
    if(bp->fd>-1) close(bp->fd);
    bp->fd=-1;
  }
}

void delbuf(struct buf_t *bp) {
  if(bp->magic!=BUFMAG) error(1,0,"delbuf: arg passed is not a buffer");
  buffreeze(bp);
  if(bp->file) unlink(bp->file);
  if(bp->p) free(bp->p);
  free(bp);
}

static void bufwrite(BUF*bp,char*buf,int n) {
 int nw,o,sz;
  if(bp->magic!=BUFMAG) error(1,0,"bufwrite: arg passed is not a buffer");
  for(o=0;o<n;o+=nw) {
    sz=(n-o<4096)?n-o:4096;
    if((nw=write(bp->fd,buf+o,sz))<0) {
      error(1,errno,"bufwrite: write error on %s",bp->file);
    }
    if(nw!=sz) error(1,0,"bufwrite: short write");
  }
}

static void buftof(struct buf_t *bp) {
  if(bp->magic!=BUFMAG) error(1,0,"buftof: arg passed is not a buffer");
  bp->file=tempnam("/tmp","lkp");
  if(!bp->file) error(1,errno,"buftof: can't make temp file");
  if((bp->fd=creat(bp->file,0600))<0) {
    error(1,0,"buf: can't create tmpfile %s",bp->file);
  }
  bufwrite(bp,bp->p,bp->pos);
  free(bp->p);
  bp->p=NULL;
}

const int FLIM=524288;

static void bufresz(struct buf_t *bp,int sz) {
 /* enlarge by at least sz bytes */
 int newsz;
  if(bp->magic!=BUFMAG) error(1,0,"bufresz: arg passed is not a buffer");
  newsz=(int)(1.4*(double)(bp->sz+sz));
  if(newsz>FLIM) buftof(bp); else {
    if(!(bp->p=realloc(bp->p,newsz))) error(1,errno,"bufresz: realloc failed");
    bp->sz=newsz;
  }
}

void addbuf(struct buf_t *bp,char*buf,int n) {
  if(bp->magic!=BUFMAG) error(1,0,"addbuf: arg passed is not a buffer");
  if(bp->frozen) error(1,0,"addbuf: trying to write frozen buffer");
  if(!bp->file&&n>bp->sz-bp->pos) bufresz(bp,n-(bp->sz-bp->pos));
  if(bp->file) {
      bufwrite(bp,buf,n);
      bp->pos+=n;
    } else {
      if(n>bp->sz-bp->pos) error(1,0,"addbuf: buffer overflow");
      memcpy(bp->p+bp->pos,buf,n);
      bp->pos+=n;
  }
}

void bufdump(struct buf_t *bp) {
 int i;
  if(bp->file) { /* buf is in file */
    printf("bufdump: buffer is in file\n");
  } else { /* buff is in mem */
    if(bp->magic!=BUFMAG) error(1,0,"bufdump: arg passed is not a buffer");
    for(i=0;i<bp->pos;i++) putchar(bp->p[i]);
  }
}

char*bufptr(struct buf_t *bp) {
  if(bp->magic!=BUFMAG) error(1,0,"bufptr: arg passed is not a buffer");
  if(bp->file) error(1,0,"bufptr: fixme: buffer is in file");
  buffreeze(bp);
  return(bp->p);
}

char*buffile(struct buf_t *bp) {
  if(bp->magic!=BUFMAG) error(1,0,"buffile: arg passed is not a buffer");
  if(!bp->file) buftof(bp);
  buffreeze(bp);
  return(bp->file);
}

int buflen(struct buf_t *bp) {
  if(bp->magic!=BUFMAG) error(1,0,"buflen: arg passed is not a buffer");
  return(bp->pos);
}

/*----------------------------------------------------------------------*/

#define UBUFMAG 0xabf04839

UBUF*ubnew(void) { /* make an ubuf with sz entries */
 UBUF*ub; int memsz,sz=5;
  if(!(ub=malloc(sizeof(UBUF)))) error(1,errno,"ubnew: out of mem");
  ub->magic=UBUFMAG;
  memsz=sz*sizeof(UBUFE);
  if(!(ub->a=malloc(memsz))) error(1,errno,"ubnew: out of mem");
  memset(ub->a,0,memsz);
  ub->sz=sz;
  ub->n=0;
  return(ub);
}

void ubdel(UBUF*ub) { /* delete ubuf */
 int i;
  if(ub->magic!=UBUFMAG) error(1,0,"ubdel: arg passed is no ubuf");
  ub->magic=0;
  for(i=0;i<ub->n;i++) {
    if(ub->a[i].url) free(ub->a[i].url);
    if(ub->a[i].host) free(ub->a[i].host);
  }
  free(ub);
}

static void ubresize(UBUF*ub) {
 int newsz,sz;
  if(ub->magic!=UBUFMAG) error(1,0,"ubresize: arg passed is no ubuf");
  sz=ub->sz;
  newsz=1.5*sz;
  if(newsz<sz+10) newsz=sz+10;
  dp(DTRC,"ubrsz: resizing from %d to %d",sz, newsz);
  if(!(ub->a=realloc(ub->a,newsz*sizeof(UBUFE)))) {
    error(1,errno,"ubresize: realloc");
  }
  ub->sz=newsz;
}

void ubdelel(UBUF*ub,int nd) { /* delete entry nd */
 int i; UBUFE*a=ub->a;
  if(ub->magic!=UBUFMAG) error(1,0,"ubdelel: arg passed is no ubuf");
  (ub->n)--;
  for(i=nd;i<ub->n;i++) a[i]=a[i+1];
}

UBUFE*ubel(UBUF*ub,int i) { /* retrieve el i */
  if(ub->magic!=UBUFMAG) error(1,0,"ubel: arg passed is no ubuf");
  if(i<0||i>ub->n) error(1,0,"ubel: ref to element %d oob (n=%d)",i,ub->n);
  return(&(ub->a)[i]);
}

void ubaddurl(UBUF*ub,char*s) { /* append entry */
 char*p;
  if(ub->magic!=UBUFMAG) error(1,0,"ubaddurl: arg passed is no ubuf");
  if(ub->n>=ub->sz) ubresize(ub);
  if(ub->n>=ub->sz) error(1,0,"ubaddurl: overflow");
  if(!(p=strdup(s))) error(1,errno,"ubaddurl: strdup out of mem");
  (ub->a)[ub->n].url=p;
  ub->n++;
}

void ubaddent(UBUF*ub,UBUFE*ube) {
  if(ub->magic!=UBUFMAG) error(1,0,"ubaddent: arg passed is no ubuf");
  if(ub->n>=ub->sz) ubresize(ub);
  if(ub->n>=ub->sz) error(1,0,"ubaddent: overflow");
  ub->a[ub->n]=*ube;
  ub->n++;
}

char*ubgeturl(UBUF*ub,int n) { /* rtn entry nr n */
  if(ub->magic!=UBUFMAG) error(1,0,"ubgeturl: arg passed is no ubuf");
  if(n<0||n>=ub->n) error(1,0,"ubgeturl: index %d out of bounds",n);
  return((ub->a)[n].url);
}

int ublen(UBUF*ub) { /* return length */
  if(ub->magic!=UBUFMAG) error(1,0,"ublen: arg passed is no ubuf");
  return(ub->n);
}

static int portcmp(const void*px,const void *py) {
 UBUFE*pa=(UBUFE*)px,*pb=(UBUFE*)py;
  return(pa->port-pb->port);
}

static int qcmp(const void*px,const void *py) {
 UBUFE*pa=(UBUFE*)px,*pb=(UBUFE*)py;
  if(pa->q<pb->q) return(-1);
  if(pa->q>pb->q) return(1);
  return(0);
}

static int dupcmp(const void*px,const void *py) {
 UBUFE*pa=(UBUFE*)px,*pb=(UBUFE*)py;
 u32 a=ntohl(pa->ip),b=ntohl(pb->ip);
  if(a<b) return(-1);
  if(a>b) return(1);
  if(pa->port<pb->port) return(-1);
  if(pa->port>pb->port) return(1);
  return(0);
}

void ubsortbyport(UBUF*ub) { /* sort ubuf by url */
  if(ub->magic!=UBUFMAG) error(1,0,"ubsortbyurl: arg passed is no ubuf");
  qsort(ub->a,ub->n,sizeof(UBUFE),portcmp);
}

void ubsortbyq(UBUF*ub) {
  if(ub->magic!=UBUFMAG) error(1,0,"ubsortbyq: arg passed is no ubuf");
  qsort(ub->a,ub->n,sizeof(UBUFE),qcmp);
}

int ubexist(UBUF*ub,u32 ip,int port) {
 int i; UBUFE*a;
  if(ub->magic!=UBUFMAG) error(1,0,"ubexist: arg passed is no ubuf");
  a=ub->a;
  for(i=0;i<ub->n;i++) {
    if(a[i].ip==ip && a[i].port==port && a[i].q!=LONG_MAX) return(1);
  }
  return(0);
}

void ubdeldups(UBUF*ub) { /* del entries with same ip/port */
 int i,port=0; u32 ip=0;
 u32 ipmax=0xffffffff;
 UBUFE*a;
  a=ub->a;
  if(ub->magic!=UBUFMAG) error(1,0,"ubdeldups: arg passed is no ubuf");
#if 0
printf("ubdeldup: n=%d\n",ub->n);
printf("-------------ubdeldup: host-dump------------------\n");
{int i;
for(i=0;i<ub->n;i++) {
 if(!a[i].ip) printf("*******\n");
 printf("%d: %s\n",i,dotquad0(a[i].ip));
}
}
printf("-------------------------------------------------\n");
#endif
  qsort(ub->a,ub->n,sizeof(UBUFE),dupcmp);
  for(i=0;i<ub->n;i++) {
    if(ip==a[i].ip&&port==a[i].port) { /* dup */
#if 0
{static c=1; printf("ubdeldup: dup %d found at pos %d: %s\n",c++,i,dotquad(ip)); }
#endif
      a[i].ip=ipmax; a[i].port=0;
      continue;
    }
    ip=a[i].ip; port=a[i].port;
  }
  qsort(ub->a,ub->n,sizeof(UBUFE),dupcmp);
  for(i=0;i<ub->n;i++) if(a[i].ip==ipmax&&a[i].port==0) break;
  ub->n=i;
/*printf("ubdeldup: n=%d\n",ub->n);*/
}

void ubtrunc(UBUF*ub, int n) {
  if(ub->magic!=UBUFMAG) error(1,0,"ubtrunc: arg passed is no ubuf");
  if(n>=0&&n<ub->n) ub->n=n;
}
